/* *************************************************************
   RWhois Software

   Copyright (c) 1994 Scott Williamson and Mark Kosters
   Copyright (c) 1996-2000 Network Solutions, Inc.

   See the file LICENSE for conditions of use and distribution.
**************************************************************** */

#include "pgp_ops.h"

#include "client_msgs.h"
#include "defines.h"
#include "fileutils.h"
#include "log.h"
#include "main_config.h"
#include "misc.h"
#include "strutil.h"

#define KEYID_FILE_TEMPLATE   "%s/keyid.%s.XXXXXX"

/*-------------------- PGP PROGRAM ENVIRONMENT ----------------*/

/* To use the pgp program we need to be able to find the executable
   the executable needs to be able to find the public and secret key
   files (PGPPATH).
 
   This does not just use the pgp-bin-path and pgp-key-path RWhois
   configuration parameters because other programs also need to use
   the PGP routines.  It also does not use putenv() since I was having
   some trouble getting the environment variable propogated to the
   shell running pgp if putenv() was called from set_key_path().
   
   IMPORTANT:  only the routines for decrypting have been modified
   to use this. */

static char *pgp_program = 0;
static char *pgp_key_path = 0;

/* get_pgp_cmd( args, nargs )
 
   Construct a PGP command line using the keyring path set with
   PGP_set_key_path(), the PGP executable set with
   PGP_set_program_name() and the specified PGP command arguments.
   The arguments will be separated by spaces.  This routine returns an
   allocated string with the command line -- the caller must free the
   string when finished. */
static char *
get_pgp_cmd( char *args[], int nargs )
{
  char *program = ( pgp_program && *pgp_program? pgp_program : "pgp" );
  char *batchmode = "+batchmode";
  char *varname = "PGPPATH=";
  char *keypath = ( pgp_key_path && *pgp_key_path ? pgp_key_path : "");
  int  len = 0;
  int  i;
  char *cmd;

  len += strlen(varname)+strlen(keypath)+1+strlen(program)+1+strlen(batchmode);
  for ( i=0; i < nargs; i++ )
  {
    len += 1 + strlen(args[i]);
  }

  cmd = xcalloc( len+1, 1 );
  if ( *keypath )
  {
    strcat( cmd, varname );
    strcat( cmd, keypath );
    strcat( cmd, " " );
  }

  strcat( cmd, program );
  strcat( cmd, " " );
  strcat( cmd, batchmode );

  for ( i=0; i < nargs; i++ )
  {
    strcat( cmd, " " );
    strcat( cmd, args[i] );
  }
  
  return cmd;
} /* get_pgp_cmd() */

int
PGP_set_program_name( progname )
  char *progname;
{
  if ( !progname || !*progname )
  {
    return FALSE;
  }
  
  if ( pgp_program )
  {
    free( pgp_program );
  }

  pgp_program = xstrdup( progname );
  return TRUE;
} /* PGP_set_program_name() */

int
PGP_set_key_path( path )
  char *path;
{
  if ( !path || !*path )
  {
    return FALSE;
  }

  if ( pgp_key_path )
  {
    free( pgp_key_path );
  }
  pgp_key_path = xstrdup( path );
  return TRUE;
} /* PGP_set_key_path() */

/*-------------------- LOCAL FUNCTIONS ------------------------*/

int 
is_id_hex(id)
  char *id;
{
  int i;
 
  for (i = 0; i < strlen(id); i++)
  {
    if (!isxdigit(id[i]))
    {
      return FALSE;
    }
  }
  return TRUE;
}


static void 
set_io_env(in, out, err)
  FILE *in;
  FILE *out;
  FILE *err;
{
  int fd[2];

  if (pipe(fd) < 0)
  {
    fprintf(stderr, "Pipe failed.");
    exit(-1);
  }

  if ((in = fdopen(fd[0], "r")) == NULL)
  {
    fprintf(stderr,"fdopen error");
    exit(1);
  }
  if ((out = fdopen(fd[1], "w")) == NULL)
  {
    fprintf(stderr,"fdopen error");
    exit(1);
  }
  if ((err = fdopen(fd[1], "w")) == NULL)
  {
    fprintf(stderr,"fdopen error");
    exit(1);
  }

  if (dup2(fd[0], 0) == -1)
  {
    fprintf(stderr,"dup error");
    exit(1);
  }
  if (dup2(fd[1], 1) == -1)
  {
    fprintf(stderr,"dup error");
    exit(1);
  }
  if (dup2(fd[1], 2) == -1)
  {
    fprintf(stderr,"dup error");
    exit(1);
  }
}


static char * 
get_userid(instr)
  char *instr;
{
  int   i;
  int   length;
  int   start;
  char  *outstr;

  length = strlen(instr);

  for (i = 0; instr[i] != '"'; i++)
  {
  }

  if (!(outstr = (char *)calloc(1, sizeof(char) * (length - i) + 5)))
  {
    exit(-1);
  }

  outstr[0] = '"';
  start = i + 1;

  for (i = start; instr[i] != '"'; i++)
  {
    outstr[(i - start) + 1] = instr[i];
  }

  outstr[(i - start) + 1] = '"';
  outstr[(i - start) + 2] = '\0';
  
  return outstr;
}



static char * 
get_keyid(instr)
  char *instr;
{
  int   i;
  int   length;
  int   start;
  char  *keyid;

  printf("%s\n", instr);
  
  for (i = 0; instr[i] != '/'; i++)
  {
  }

  start = i + 1;

  for (i = start; instr[i] != ' '; i++)
  {
  }

  length = i - start;
  
  if (!(keyid = (char *)calloc(1, sizeof(char) * (length) + 5)))
  {
    exit(-1);
  }

  for (i = start; instr[i] != ' '; i++)
  {
    keyid[i - start] = instr[i];
  }

  keyid[i - start] = '\0';

  return keyid;
}


static void
decrypt_message(input_file, output_file, pub_ring, keyid)
  char *input_file;
  char *output_file;
  char *pub_ring;
  char *keyid;
{
  FILE  *pgpfp1;
  FILE  *pgpfp2;

  FILE  *in     = NULL;
  FILE  *out    = NULL;
  FILE  *err    = NULL;

  FILE  *fp;

  char  buf1[1024];
  char  buf2[1024];
  char  *userid = NULL;
  char  *key = NULL;
  char  keyids[MAX_LINE];
  char  *keyid_file = tempnam(".", "pgp");
  int   childpid;
  char  *newcmd;
  char  *args[4];

  bzero(keyids, sizeof(keyids));
 
  if ((childpid = fork()) < 0)
  {
    exit(-1);
  }
  if (childpid == 0)
  {
    set_io_env(in, out, err);

    args[0] = input_file;
    args[1] = "-o";
    args[2] = output_file;
    if ( !(newcmd = get_pgp_cmd(args,3)) )
    {
      exit (-1);
    }

    if ((pgpfp1 = popen(newcmd, "r")) == NULL)
    {
      exit(-1);
    }
    while (fgets(buf1, 1024, pgpfp1) != NULL)
    {
      if (strncmp(buf1, "Good", 4) != 0)
      {
        continue;
      }
      userid = get_userid(buf1);

      /* extract the keyID of this user */
      
      free(newcmd);

      args[0] = "-kv";
      args[1] = userid;
      args[2] = (pub_ring && *pub_ring ? pub_ring : 0);
      if ( !(newcmd = get_pgp_cmd(args, (args[2] ? 3 : 2))) )
      {
        exit (-1);
      }

      if ((pgpfp2 = popen(newcmd, "r")) == NULL)
      {
        exit(-1);
      }
      while (fgets(buf2, 1024, pgpfp2) != NULL)
      {
        if (strncmp(buf2, "pub", 3) != 0)
        {
          continue;
        }
        key = get_keyid(buf2);
        if(keyids[0] == '\0')
        {
          sprintf(keyids, "%s ", key);
        }
        else 
        {
          strcat(keyids, key);
          strcat(keyids, " ");
        }
      }
      pclose(pgpfp2);
    }
    pclose(pgpfp1);
    if ((fp = fopen(keyid_file, "w")) == NULL)
    {
      exit(-1);
    }
    fprintf(fp, "%s", keyids);
    fclose(fp);
        
    fclose(in);
    fclose(out);
    fclose(err);
    
    free(userid);
    free(key);
    free(newcmd);
    free(keyid_file);

    exit(0);
  }
  else 
  {
    wait((int *)NULL);
    if (!(fp = fopen(keyid_file, "r")))
    {
      fprintf(stderr, "Error opening file\n");
      exit(-1);
    }
    fgets(keyids, MAX_LINE -1, fp);
    strcpy(keyid, keyids);
    fclose(fp);
    unlink(keyid_file);
    free( keyid_file );
  }
}


int 
encrypt_msg(cmd, input_file, out_file, her_id, my_id, passphrase)  
  char *cmd;
  char *input_file;
  char *out_file;
  char *her_id;
  char *my_id;
  char *passphrase;
{
  int pw_fd[2];
 
  int childpid;
  FILE *pgpfp;
  FILE *pw_file;
  FILE *in      = NULL;
  FILE *out     = NULL;
  FILE *err     = NULL;
  FILE *fp;
  char *pgpcmd;
  char pgpenv[25];
 
  if (passphrase && *passphrase)
  {
    if (pipe(pw_fd) < 0)
    {
      fprintf(stderr, "Pipe failed\n");
      exit(-1);
    }
  } 
  if ((childpid = fork()) < 0)
  {
    exit(-1);
  }
  if (childpid == 0)
  {
    set_io_env(in, out, err);

    if (passphrase && *passphrase)
    {
      sprintf(pgpenv, "PGPPASSFD=%d", pw_fd[0]);
      putenv(pgpenv);

      if (!her_id || !*her_id)
      {
        her_id = "";
      }
      if (!(pgpcmd = (char *)calloc(1, sizeof(char) * 
                                   (strlen(cmd) + strlen(input_file) + 
                                    strlen(her_id) + strlen(my_id) +
                                    strlen(out_file)))))
      {
        exit(-1);
      }
      sprintf(pgpcmd, "%s %s %s -u %s -o %s", cmd, input_file, her_id,
                       my_id, out_file);
    } 
    else
    {
      if (strlen(her_id) == 0)
      {
        exit(-1);
      }
      if (!(pgpcmd = (char *)calloc(1, sizeof(char) * 
                                   (strlen(cmd) + strlen(input_file) + 
                                    strlen(her_id) + strlen(out_file)))))
      {
        exit(-1);
      }
      sprintf(pgpcmd, "%s %s %s -o %s", cmd, input_file, her_id, out_file);
    } 
      
    if (!(pgpfp = popen(pgpcmd, "w")))
    {
      exit(-1);
    }
    pclose(pgpfp);
  }
  if (passphrase && *passphrase)
  {
    pw_file = fdopen(pw_fd[1], "w");
    fprintf(pw_file, "%s\n", passphrase);
    fflush(pw_file);
    fclose(pw_file);
  }
  wait((int *)NULL);
  if (!(fp = fopen(out_file, "r")))
  {
    return FALSE;
  }
  return TRUE;
}


int 
add_key(keyfile)
  char *keyfile;
{
  FILE *in      = NULL;
  FILE *out     = NULL;
  FILE *err     = NULL;
  FILE *pgpfp   = NULL;
  
  int childpid;

  char *cmd = "pgp +batchmode -ka";
  char *pgpcmd = NULL;

  set_io_env(in, out, err);  

  if ((childpid = fork()) < 0)
  {
    fclose(in);
    fclose(out);
    fclose(err);
    fprintf(stderr, "Fork error\n");
    return FALSE;
  }
  if (childpid == 0)
  {
    if (!(pgpcmd = (char *)calloc(1, sizeof(char) * (strlen(cmd) +
                                   strlen(keyfile) + 10))))
    {
      fprintf(stderr, "Memory allocation problem\n");
      return FALSE;
    }

     sprintf(pgpcmd, "%s %s", cmd, keyfile); 
    if (!(pgpfp = popen(pgpcmd, "w"))) 
    {
      fprintf(stderr, "popen error\n");
      return FALSE;
    }
    pclose(pgpfp);
  }

  wait((int *)NULL);
  /* fclose(in); */
  /* fclose(out); */
  /* fclose(err); */
  /*
  while(fgets(buf, 1024, in))
  {
    if (strcmp(buf, "Keyring add error.") == 0)
    {
      return FALSE;
    }
  }
  
  printf("I am here\n");
  */
  return TRUE;
}

  


/*---------------- PUBLIC FUNCTIONS ------------------------*/

int
PGP_decrypt_msg(input_file, output_file, keyid, pub_ring)
  char *input_file;
  char *output_file;
  char *keyid;
  char *pub_ring;
{

  if (!input_file || !*input_file)
  {
/*   log(L_LOG_ERR, UNKNOWN, "decrypt_pgp_msg: input file name required"); 
*/
    return FALSE;
  }

  if (!output_file || !*output_file)
  {
/*   log(L_LOG_ERR, UNKNOWN, "decrypt_pgp_msg: output file name required"); 
*/
    return FALSE;
  }

  *keyid = '\0';
  decrypt_message(input_file, output_file, pub_ring, keyid);
  return *keyid != '\0';
}




int
PGP_verify_keyid(pub_ring, keyid)
  char *pub_ring;
  char *keyid;
{
  FILE *fp;
  FILE *in  = NULL;
  FILE *out = NULL;
  FILE *err = NULL;
  FILE *pgpfp;

  char buf1[1050];
  int childpid;
  char *tmpfile;

  char  buf[MAX_LINE];
  int   rval = FALSE;

  if (!keyid || !*keyid)
  {
    return(rval);
  }

  tmpfile = tempnam(".", "pgptmp");

  if ((childpid = fork()) < 0)
  {
    exit(-1);
  }
  if (childpid == 0)
  {
    set_io_env(in, out, err);

    bzero((char *)buf, sizeof(buf));

    if (pub_ring)
    { 
      if (is_id_hex(keyid))
      {
        sprintf(buf, "pgp -kv 0x%s %s ", keyid, pub_ring);
      }
      else
      {
        sprintf(buf, "pgp -kv %s %s", keyid, pub_ring);
      }
    }
    else
    {
      if (is_id_hex(keyid))
      {
        sprintf(buf, "pgp -kv 0x%s", keyid);
      }
      else
      {
        sprintf(buf, "pgp -kv %s", keyid);
      }
    }
      
    if ((pgpfp = popen(buf, "r")) == NULL)
    {
      exit(-1);
    }
    while (fgets(buf1, 1024, pgpfp)!= NULL)
    {
      if (!strncmp(buf1, "1 matching key found", 20))
      {
        rval = TRUE;
        break;
      }
    }
    fclose(pgpfp);
    fp = fopen(tmpfile, "w");
    fprintf(fp,"%d\n", rval);
    fclose(fp);
    fclose(in);
    fclose(err);
    exit(0);
  }
  else
  {
    wait((int *)NULL);
    fp = fopen(tmpfile, "r");
    fscanf(fp, "%d", &rval);
    fclose(fp);
    unlink(tmpfile);
    return rval;
  }

  return FALSE;
}



int 
PGP_sign_msg(msg_file, out_file, my_id, passphrase)
  char *msg_file;
  char *out_file;
  char *my_id;
  char *passphrase;
{
  FILE *fp;
  char *cmd = "pgp +batchmode -sa";

  if (!msg_file || !*msg_file || !(fp = fopen(msg_file, "r")))
  {
    fprintf(stderr, "Input file not found\n");
    return FALSE;
  }
  fclose(fp);

  if (encrypt_msg(cmd, msg_file, out_file, "", my_id, passphrase))
  {
    return TRUE;
  }
  else
  {
    return FALSE;
  }
}


int 
PGP_encrypt_msg(msg_file, out_file, keyid)
  char *msg_file;
  char *out_file;
  char *keyid;
{
  FILE *fp;
  char *cmd = "pgp +batchmode -ea";

  if (!msg_file || !*msg_file ||
      !keyid || !*keyid)
  {
    return FALSE;
  }

  if (!(fp = fopen(msg_file, "r")))
  {
    fprintf(stderr, "Input file not found\n");
    return FALSE;
  }
  fclose(fp);

  if (PGP_verify_keyid("", keyid) != TRUE)
  {
    fprintf(stderr, "Key id not found in keyring\n");
    return FALSE;
  }

  if (encrypt_msg(cmd, msg_file, out_file, keyid, "", ""))
  {
    return TRUE;
  }
  else
  {
    return FALSE;
  }
}


int 
PGP_sign_encrypt_msg(msg_file, out_file, keyid, my_id, passphrase)
  char *msg_file;
  char *out_file;
  char *keyid;
  char *my_id;
  char *passphrase;
{
  FILE *fp;
  char *cmd = "pgp +batchmode -esa";

  if (!msg_file || !*msg_file ||
      !keyid || !*keyid)
  {
    return FALSE;
  }

  if (!(fp = fopen(msg_file, "r")))
  {
    fprintf(stderr, "Input file not found\n");
    return FALSE;
  }
  fclose(fp);

  if (PGP_verify_keyid("", keyid) != TRUE)
  {
    fprintf(stderr, "Key id not found in keyring\n");
    return FALSE;
  }

  if (encrypt_msg(cmd, msg_file, out_file, keyid, my_id, passphrase))
  {
    return TRUE;
  }
  else
  {
    return FALSE;
  }
}


int PGP_add_key(char *key)
{
  char *keyfile = tempnam(".", "key");
  FILE *keyfp;

  if (!key || !*key)
  {
    return FALSE;
  }
  if (!(strstr(key, "BEGIN PGP PUBLIC KEY BLOCK")) ||
      !(strstr(key, "END PGP PUBLIC KEY BLOCK")))
  {
    fprintf(stderr, "Not valid data or data corrupted\n");
    return FALSE;
  }
  /*
  if (!(end = strstr(key, "-X-pgp off")))
  {
     return FALSE; 
  }
  *end = '\0';
  */

  keyfp = fopen(keyfile, "w");
  fwrite(key, strlen(key), 1, keyfp);
  fclose(keyfp);

  if (!add_key(keyfile))
  { 
    return FALSE;
  }
  unlink(keyfile);
  return TRUE;
}

/*

main(int argc, char **argv)
{
  char *input_file;
  char *output_file;
  FILE *in;

   char keyid[MAX_LINE]; 
  char *key;
 
  keyid[0] = '\0';
 
  if (argc < 3)
  {
    fprintf(stderr, "usage: pgptest <encrypted file name> <plaintext file>\n");
    exit(0);
  }
 
  if (!(input_file = (char *) calloc(1, sizeof(char) *
    (strlen(argv[1]) + 10))))
  {
    fprintf(stderr, "Memory allocation problem");
    exit(1);
  }
 
  strcpy(input_file, argv[1]);
 
  if (!(output_file = (char *) calloc(1, sizeof(char) *
    (strlen(argv[2]) + 10))))
  {
    fprintf(stderr, "Memory allocation problem");
    exit(1);
  }
  strcpy(output_file, argv[2]);

  if (!(PGP_decrypt_msg(input_file, output_file, keyid, "")))
  {
    fprintf(stderr, "Cannot decrpypt message");
  }
  printf("KeyID = %s\n", keyid);
  if (!(PGP_verify_keyid("", keyid)))
  {
    printf("booooo\n");
  }
  else
  {
    printf("Key found\n");
  }

  unlink(input_file);
  free(input_file);


  if ((input_file = PGP_encrypt_msg(output_file, "", keyid, 
    PGP_ENCRYPT)) == NULL)
  {
    printf("Error in PGP_encrypt_msg\n");
    exit(1);
  }
  printf("Output file is %s\n", input_file);
  
  if (!(PGP_verify_keyid("", argv[1])))
  {
    printf("Verification failed\n");
  }
  printf("success");


  in = fopen(argv[1], "r");
  key = (char *)calloc(1, sizeof(char) * 1050);
  while(!(feof(in)))
  {
    fread(key, 1024, 1, in);
  }
  printf("%s\n", key);

  PGP_add_key(key);
  
}
*/
